﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO.Ports;
using System.IO;
using System.Windows.Forms.DataVisualization.Charting;
/*
Форматирование данных для программы Arduino reader ( дальше AR )

Cтроки которые сможет обработать программа чтения порта

каждая из строк состоит из частей
Примеры строк:
   строка заголовка разделяется запятой
      #graph,1,5,p,5.0,R1,0,255,0  
   строка графика
      g:1903000:1000:41:104:665:481:860
   строка  сообщения
      m:SHIFT0 was changed to 10

частей должно быть не менее 2
первая часть - опознавательная, что будет в этой строке, вторая и остальные содержательные
может быть:
# - заголовочное описание. В нем содержится информация о формате каждой из строк данных
   Заголовочные описания должны быть отправлены самыми первыми. Пока PC не получит заголовки, он не сможет расшифровать строки данных.
   Возможные заголовки 
   #text - описывает данные для ОДНОГО текстового поля программы AR
      следующими частями в заголовке #text должны быть
	 - номер текстового поля в программе AR, целое число 0..4 
	 - номер части строки графика (g) в выдаче COM порта Arduino, целое число ( 1..x3 )
	 - заголовок, который будет присвоен текстовому полю в AR, строка английскими буквами
      Примеры
	 #text,0,2,Period
	    в текстовое поле №0 программы AR будут попадать данные из второй части строк g: 
	    например, для строки g:1903000:1000:41:104:665:481:860, в текстовое поле №0 попадет 1000, так как части нумеруются с 0
	    поле №0 в программе AR будет подписано "Period"
	 #text,2,8,Measure time
	    в текстовое поле №2 программы AR будут попадать данные из восьмой части строк g: 
	    например, для строки g:11131900:500:492:56:494:778:387:442:520:4, в текстовое поле №2 попадет 442, так как части нумеруются с 0
	    поле №0 в программе AR будет подписано "Measure time"
   Данные для текстовых полей могут быть любыми, но если данные будут числовыми, AR проведет нехитрые вычисления: отследит min max и 
   посчитает среднее значение за все время. Поле будет выглядеть так: 442 430<444<460

   #graph - описывает данные для ОДНОГО графика программы AR
      следующими частями в заголовке #graph должны быть
	 - номер графика в программе AR, для которого описывается точка, целое число 0..4 
	 - номер части строки графика (g) в выдаче COM порта Arduino которую AR использует как данные точки, целое число ( 1..x3 )
	 - символ вертикальной шкалы к которой относится график, p или s, от англ. Primary Secondary. p - левая шкала, s - правая
	 - максимальное значение для перерасчёта значения точки, double. Сам Arduino должен слать привычные для себя данные, 0..1023,
	       а AR пересчитает точки в реальные значения измерений по формуле 
	       значение = максимальное_значение * данные_точки / 1024
	 - название графика, строка английскими буквами. AR добавит к этому названию максимальное значение для простоты восприятия
	 - последние 3 целых (0.255) - это цвет графика в формате RGB 
      Примеры
	 #graph,0,3,p,5.0,R0,255,0,0
	    График №0 будет строиться из значений третьей части строк g:, данным 1023 будет соответствовать метка 5.0 на левой шкале.
	    График будет красного цвета и подписан "R0, 5.0".
	 #graph,2,6,s,12.0,R2,0,0,255
	    График №2 будет строиться из значений шестой части строк g:, данным 1023 будет соответствовать метка 12.0 на правой шкале.
	    График будет синего цвета и подписан "R2, 12.0".
	    
   #SysMessagesDone - сообщает AR, что описательная часть закончилась, дальше идут строки данных 
	    
Примечания
Первыми данными в порт стоит посылать несколько пустых строк, потому как на момент первого чтении, там может быть всякий мусор, оставшийся от предыдущей сессии.
Этот мусор не даст распознать первую строку заголовков.
Не забывайте завершать заголовок строкой #SysMessagesDone. Пока её нет, AR ничего не делает.

*/
namespace arduino
{
    public partial class Form1 : Form
    {
        bool disablePort     = false;
        bool SysMessagesDone = false;
        string path;
        double[] GraphsHigest= new double[5] {0,0,0,0,0};
        int[] GraphsPos = new int[5] { -1, -1, -1, -1, -1 };
        int[] TextsPos = new int[5] { -1, -1, -1, -1, -1 };
        double[] TextMin = new double[5] { 99999999, 99999999, 99999999, 99999999, 99999999 };
        double[] TextMax = new double[5] {0,0,0,0,0};
        double[] TextAcc = new double[5] {0,0,0,0,0};
        int[] TextCnt    = new int[5] { 0, 0, 0, 0, 0 };
        TextBox[] TextBoxes = new TextBox[5];


        private void Application_Idle(Object sender, EventArgs e)
        {
            btnSave.Enabled  = (chart1.Series[0].Enabled || chart1.Series[1].Enabled || chart1.Series[2].Enabled || chart1.Series[3].Enabled || chart1.Series[4].Enabled);
        }
        public Form1()
        {
            InitializeComponent();
            Application.Idle += new EventHandler(Application_Idle);

            TextBoxes[0]=texts0; TextBoxes[1]=texts1; TextBoxes[2]=texts2; TextBoxes[3]=texts3; TextBoxes[4]=texts4;
            resetGraphs();

            notifyIcon1.Visible = false;
            this.notifyIcon1.MouseClick += new MouseEventHandler(notifyIcon1_MouseClick);
            this.Resize += new System.EventHandler(this.Form1_Resize);

            string[] ports = SerialPort.GetPortNames();
            if (ports.GetLength(0) > 0)
            {
                cbPorts.Items.Add("Select Port");
                for (int i = 0; i < ports.GetLength(0); i++)
                {
                    cbPorts.Items.Add(ports[i]);
                }
            }
            else
            {
                cbPorts.Items.Add("No Ports found");
            }
            if (Properties.Settings.Default.lastCom != null && Properties.Settings.Default.lastCom.StartsWith("COM"))
                cbPorts.SelectedIndex = Math.Max(0,cbPorts.Items.IndexOf(Properties.Settings.Default.lastCom));
            else
                cbPorts.SelectedIndex = 0;

            if (Properties.Settings.Default.lastComSpeed != null && cbPortSpeed.Items.IndexOf( Properties.Settings.Default.lastComSpeed )>=0 )
                cbPortSpeed.SelectedIndex = cbPortSpeed.Items.IndexOf( Properties.Settings.Default.lastComSpeed ) ;
            else
                cbPorts.SelectedIndex = cbPortSpeed.Items.IndexOf("9600");
            udGWidth.Value = Math.Min(3, Math.Max(1, Properties.Settings.Default.udGWidth));
            

            
            // Открываем порт, и задаем скорость в 9600 бод
            serialPort1.BaudRate = 9600;
            serialPort1.DtrEnable = true;
            serialPort1.DataReceived += serialPort1_DataReceived;

            chart1.ChartAreas[0].CursorX.LineColor = Color.Yellow;
            chart1.ChartAreas[0].CursorX.IsUserEnabled = true;
            chart1.ChartAreas[0].CursorX.IsUserSelectionEnabled = true;
            chart1.ChartAreas[0].AxisX.ScaleView.Zoomable = true;
            chart1.ChartAreas[0].AxisX.ScrollBar.IsPositionedInside = true;

            chart1.ChartAreas[0].CursorY.LineColor = Color.Yellow;
            chart1.ChartAreas[0].CursorY.IsUserEnabled = true;
            chart1.ChartAreas[0].CursorY.IsUserSelectionEnabled = true;
            chart1.ChartAreas[0].AxisY.ScaleView.Zoomable = true;
            chart1.ChartAreas[0].AxisY.ScrollBar.IsPositionedInside = true;

            for (int i = 0; i < chart1.Series.Count; i++)
            {
                chart1.Series[i].ToolTip = "#VAL";
                chart1.Series[i].BorderWidth = (int)udGWidth.Value;
            }

//            chart1.ChartAreas[0].AxisX.IntervalAutoMode = IntervalAutoMode.VariableCount;
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (serialPort1.IsOpen )
            {
                disablePort = true;
                serialPort1.Close();
            }
        }
        private void Form1_Resize(object sender, EventArgs e)
        {
            if (this.WindowState == FormWindowState.Minimized)
            {
                this.Hide();
                notifyIcon1.Visible = true;
            }
        }

        //****** поток COM порта
        private void serialPort1_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            if ( disablePort || !serialPort1.IsOpen ) return;
            try
            {
                string data = serialPort1.ReadLine();
                this.BeginInvoke(new LineReceivedEvent(LineReceived), data.Trim());
            }
            catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        //***** запись данных от COM
        private delegate void LineReceivedEvent(string data);
        private void LineReceived(string data)
        {
            DateTime date = DateTime.Now;
            textBox1.Text = data;
            if (data.Length > 0)
            {
                if ( !data.StartsWith("#") )
                    data = date.ToString().Replace(":", "|") + ":" + data;
                // Создание файла и запись в него
                using (StreamWriter sw = File.AppendText(path))
                {
                    sw.WriteLine( data );
                }
            }
            showData(data);
        }

        void decodeInfo(string data) {
            //   #text,0,1,Period = #text + номер текстового поля 0..4 + на каком месте в строке ( с 0 ) + подпись текстового поля
            //   #graph,1,4,p,5.0,R1,0,255,0 = #graph + №графика ( chart1.Series[ =i ]) + на каком месте в строке ( с 0 ) + шкала ( p-левая, s-правая ) + макс значение + название графика + RGB цвета графика
            int i, j, k, l;
            double d;

            DateTime date = DateTime.Now;
            listBox1.SelectedIndex = listBox1.Items.Add(date.ToString() + ": " + data);

            string[] s=data.Split(new Char[] { ',' });
            switch ( s[0].ToLower() ) {
                case "#text": 
                    if(s.Length==4){ 
                        i = Convert.ToInt16(s[1]);
                        j = Convert.ToInt16(s[2]);
                        TextsPos[i] = j + 1; // потому как добавляем для каждой строки дату нулевым членом
                        switch (i) { 
                            case 0: labels0.Text = s[3]; break;
                            case 1: labels1.Text = s[3]; break;
                            case 2: labels2.Text = s[3]; break;
                            case 3: labels3.Text = s[3]; break;
                            case 4: labels4.Text = s[3]; break;
                        }
                    }; 
                break;
                case "#graph": 
                    if (s.Length == 9) { 
                        i = Convert.ToInt16(s[1]);
                        j = Convert.ToInt16(s[2]);
                        try
                        {
                            d = Convert.ToDouble(s[4].Replace(".", ","));
                        }
                        catch {
                            d = Convert.ToDouble(s[4]);
                        }
                        GraphsHigest[i] = d;
                        GraphsPos[i] = j + 1; // потому как добавляем для каждой строки дату нулевым членом
                        j = Convert.ToInt16( s[6] );
                        k = Convert.ToInt16( s[7] );
                        l = Convert.ToInt16( s[8] );
                        chart1.Series[i].Enabled=true;
                        chart1.Series[i].Color = Color.FromArgb(j, k, l);
                        chart1.Series[i].LegendText = s[5]+", "+s[4];
                        chart1.Series[i].ToolTip = s[5] + ": #VAL";
                        chart1.Series[i].YAxisType = (s[3] == "s") ? System.Windows.Forms.DataVisualization.Charting.AxisType.Secondary : System.Windows.Forms.DataVisualization.Charting.AxisType.Primary;
                    }; 
                break;
                case "#sysmessagesdone": SysMessagesDone = true; break;
            }
        }
        
        void showData(string data) {
            if (data.StartsWith("#"))
            {
                decodeInfo(data);
            }
            else
            {
                if (SysMessagesDone) // начальные строки с расшифровкой данных обработаны
                {
                    string[] p = data.Split(new Char[] { ':' });
                    string date = p[0].Replace("|", ":").Replace(" ", "\n"); // дата строки
                    switch(p[1]){
                        case "g": // строка графика
                        {
                            for (int i = 0; i < 5; i++)
                            {
                                if (TextsPos[i] >= 0)
                                {
                                    try
                                    {
                                        int intp = Convert.ToInt32(p[TextsPos[i]]); // если получили нечто неконвертируемое в целое, 
                                        if (intp < TextMin[i]) TextMin[i] = intp;
                                        if (intp > TextMax[i]) TextMax[i] = intp;
                                        TextAcc[i] += intp;
                                        TextCnt[i]++;
                                        int avg = Convert.ToInt32(TextAcc[i] / TextCnt[i]);
                                        TextBoxes[i].Text = p[TextsPos[i]] + (TextMin[i] != TextMax[i] ? " " + TextMin[i].ToString() + "<" + avg.ToString() + "<" + TextMax[i].ToString() : "");
                                    }catch( Exception ){
                                        TextBoxes[i].Text = p[TextsPos[i]]; // то просто выводим без изменений
                                    }
                                }
                            }

                            for (int i = 0; i < 5; i++)
                            {
                                if (chart1.Series[i].Enabled && GraphsPos[i] > 0 && GraphsHigest[i] > 0)
                                {
                                    double value =  Convert.ToDouble(p[GraphsPos[i]]) / 1023 * GraphsHigest[i];
                                    value = Math.Truncate(value * 1000) / 1000;
                                    chart1.Series[i].LegendText = chart1.Series[i].LegendText.Substring(0, chart1.Series[i].LegendText.IndexOf(",")+1) + value.ToString();
                                    chart1.Series[i].Points.AddXY(date, value);
                                }
                            }
                        }
                        break;
                        case "o": // строка осла
                        {

                        }
                        break;
                        case "m": // строка сообщения
                        {
                            string msg = p[2];
                            for (int i = 3; i < p.Length; i++) msg += ", " + p[i];
                            listBox1.SelectedIndex = listBox1.Items.Add(date.Replace("\n", " ") + ": " + msg);
                        }
                        break;
                        case "c": // эхо консольной команда
                        {
                            string msg = p[2];
                            listBox1.SelectedIndex = listBox1.Items.Add(date.Replace("\n", " ") + ": cmd: " + msg);
                        }
                        break;

                    }
                }
            }
        }


        private void btnFiles_Click(object sender, EventArgs e)
        {
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                string inp;
                resetGraphs();
                string fn= openFileDialog1.FileName.Substring( openFileDialog1.FileName.LastIndexOf('\\')+1 );
                this.Text = "Arduino reader чтение из [" + fn + "]";
                StreamReader streamReader = new StreamReader(openFileDialog1.FileName);
                int lnCnt=0, dropLines = Convert.ToInt32(tbDropLines.Text);
                lnCnt = 0;
                while (!streamReader.EndOfStream)
                {
                    inp = streamReader.ReadLine();
                    if (inp.StartsWith("#"))
                    {
                        showData(inp);
                    }
                    else
                    {
                        if ( lnCnt==0 )
                        {
                            showData(inp);
                            lnCnt= dropLines+1;
                        }
                        if (lnCnt > 0) lnCnt--;
                    }
                }
                streamReader.Close();
            } 
        }

        private void notifyIcon1_MouseClick(object sender, MouseEventArgs e)
        {
            this.Show();
            this.WindowState = FormWindowState.Normal;
            notifyIcon1.Visible = false;
        }

        private void resetGraphs()
        {
            SysMessagesDone = false;
            for (int i = 0; i < 5; i++)
            {
                chart1.Series[i].Points.Clear();
                chart1.Series[i].Enabled = false;
                GraphsHigest[i] = 0;
                GraphsPos[i] = TextsPos[i] = -1;
                TextBoxes[i].Text = "";
                TextMin[i] = 99999999;
                TextMax[i]=TextAcc[i]=TextCnt[i]=0;
            }
            listBox1.Items.Clear();
            labels0.Text = labels1.Text = labels2.Text = labels3.Text = labels4.Text = "x3";
        }

        private void btnPort_Click(object sender, EventArgs e)
        {
            if (serialPort1.IsOpen)
            {
                serialPort1.Close();
                this.Text="Arduino reader";
                cbPorts_SelectedValueChanged(sender, e);
                tbTextIn.Enabled = false;
            }
            else
            {
                try
                {
                    serialPort1.BaudRate = Convert.ToInt32( cbPortSpeed.SelectedItem.ToString() );
                    serialPort1.Open();
                    resetGraphs();
                    path = serialPort1.PortName + "-" + DateTime.Now.ToString("yy.MM.dd HH.mm.ss") + "-arduino.txt";
                    this.Text = "Arduino reader запись в [" + path + "]";
                    cbPortSpeed.Enabled = cbPorts.Enabled = false;
                    tbTextIn.Enabled = true;
                    btnPort.Text = "Закрыть";
                    Properties.Settings.Default.lastCom      = serialPort1.PortName;
                    Properties.Settings.Default.lastComSpeed = serialPort1.BaudRate.ToString();
                    Properties.Settings.Default.Save();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void cbPorts_SelectedValueChanged(object sender, EventArgs e)
        {
            if (cbPorts.SelectedItem != null && cbPorts.SelectedItem.ToString().StartsWith("COM"))
            {
                serialPort1.PortName = cbPorts.SelectedItem.ToString();
                btnPort.Text = "Открыть";
                btnPort.Enabled = true;
                cbPortSpeed.Enabled=cbPorts.Enabled = true;
            }
            else
            {
                btnPort.Enabled = false;
            }
        }

        private void cbPorts_EnabledChanged(object sender, EventArgs e)
        {
            lbDropLines.Enabled = tbDropLines.Enabled = btnFiles.Enabled = cbPorts.Enabled;
        }

        private void btnSave_Click(object sender, EventArgs e)
        {
            // Create a new save file dialog
            SaveFileDialog saveFileDialog1 = new SaveFileDialog();

            // Sets the current file name filter string, which determines 
            // the choices that appear in the "Save as file type" or 
            // "Files of type" box in the dialog box.
            saveFileDialog1.Filter = "Bitmap (*.bmp)|*.bmp|JPEG (*.jpg)|*.jpg|PNG (*.png)|*.png|GIF (*.gif)|*.gif";
            saveFileDialog1.FilterIndex = 3;
            saveFileDialog1.RestoreDirectory = true;
            saveFileDialog1.FileName = DateTime.Now.ToString("yy.MM.dd HH.mm.ss");

            // Set image file format
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ChartImageFormat format = ChartImageFormat.Bmp;

                if (saveFileDialog1.FileName.EndsWith("bmp"))
                {
                    format = ChartImageFormat.Bmp;
                }
                else if (saveFileDialog1.FileName.EndsWith("jpg"))
                {
                    format = ChartImageFormat.Jpeg;
                }
                else if (saveFileDialog1.FileName.EndsWith("gif"))
                {
                    format = ChartImageFormat.Gif;
                }
                else if (saveFileDialog1.FileName.EndsWith("png"))
                {
                    format = ChartImageFormat.Png;
                }

                // Save image
                chart1.SaveImage(saveFileDialog1.FileName, format);
            }
        }

        private void tbDropLines_KeyPress(object sender, System.Windows.Forms.KeyPressEventArgs e)
        {
            if (e.KeyChar >= '0' && e.KeyChar <= '9' || e.KeyChar == 8 || e.KeyChar == 9 ) return;
            e.Handled = true;
        }

        private void tbDropLines_Leave(object sender, EventArgs e)
        {
            if (tbDropLines.Text == "") tbDropLines.Text = "0";
        }

        private void btnHelp_Click(object sender, EventArgs e)
        {
            Form2 form2 = new Form2();
            form2.StartPosition = FormStartPosition.CenterScreen;
            form2.Show();
        }

        private void udGWidth_ValueChanged(object sender, EventArgs e)
        {
            for (int i = 0; i < chart1.Series.Count; i++) chart1.Series[i].BorderWidth = (int)udGWidth.Value;
            Properties.Settings.Default.udGWidth = (int)udGWidth.Value;
            Properties.Settings.Default.Save();
        }

        private void tbTextIn_KeyPress(object sender, KeyPressEventArgs e)
        {
            if ( e.KeyChar == (char)13 && tbTextIn.Text.Length > 0) {
                tbTextIn.Text=tbTextIn.Text.Trim();
                if (tbTextIn.Text.StartsWith("c:", StringComparison.OrdinalIgnoreCase ))
                {
                    string cmd = tbTextIn.Text.Substring(2).Trim();
                    serialPort1.WriteLine(cmd); // в консоль не пишем, ардуино должен эхо дать
                    
                }
                else
                {
                    DateTime date = DateTime.Now;
                    string data = ":m:hands: " + tbTextIn.Text;
                    listBox1.SelectedIndex = listBox1.Items.Add(date.ToString() + ": hands: " + tbTextIn.Text);
                    data = date.ToString().Replace(":", "|") + data;
                    using (StreamWriter sw = File.AppendText(path))
                    {
                        sw.WriteLine(data);
                    }
                }
                tbTextIn.Text = "";
                e.Handled = true;
            }
        }
    }
}
